import threading
from subprocess import Popen, PIPE, TimeoutExpired
import select
import logging


logger = logging.getLogger('periodtask.process_thread')


def _parse(head_tail):
    if head_tail is None:
        return None, None, None
    if isinstance(head_tail, int):
        return head_tail, head_tail, 2 * head_tail
    try:
        mx = head_tail[0] + head_tail[1]
    except TypeError:
        mx = None
    return head_tail[0], head_tail[1], mx


def parse_max_lines(max_lines):
    if max_lines is None:
        return None, None, None, None, None, None
    if isinstance(max_lines, int):
        return (
            max_lines, max_lines, 2 * max_lines,
            max_lines, max_lines, 2 * max_lines
        )
    (stdout_max_lines, stderr_max_lines) = max_lines
    oh, ot, om = _parse(stdout_max_lines)
    eh, et, em = _parse(stderr_max_lines)
    return oh, ot, om, eh, et, em


class ProcessThread(threading.Thread):
    def __init__(
        self, task_name, command, stop_signal, wait_timeout,
        formatted_sec, max_lines,
        stdout_logger, stdout_level, stderr_logger, stderr_level,
        cwd
    ):
        self.task_name = task_name
        self.command = command
        self.stop_signal = stop_signal
        self.wait_timeout = wait_timeout
        self.formatted_sec = formatted_sec
        self.max_lines = parse_max_lines(max_lines)
        self.stdout_logger = stdout_logger
        self.stdout_level = stdout_level or logging.INFO
        self.stderr_logger = stderr_logger
        self.stderr_level = stderr_level or logging.INFO
        self.cwd = cwd

        self.stdout_head = []
        self.stdout_tail = []
        self.stderr_head = []
        self.stderr_tail = []

        self.returncode = None
        self.proc = None
        self.lock = threading.Lock()
        super(ProcessThread, self).__init__()

    def lines(self, head, tail):
        with self.lock:
            if tail:
                return (
                    '\n'.join(head) +
                    '\n...\n' +
                    '\n'.join(tail)
                )
            else:
                return '\n'.join(head)

    def get_stdout_head(self):
        with self.lock:
            return [x for x in self.stdout_head]

    def get_stdout_tail(self):
        with self.lock:
            return [x for x in self.stdout_tail]

    def get_stderr_head(self):
        with self.lock:
            return [x for x in self.stderr_head]

    def get_stderr_tail(self):
        with self.lock:
            return [x for x in self.stderr_tail]

    @property
    def stdout_lines(self):
        return self.lines(self.stdout_head, self.stdout_tail)

    @property
    def stderr_lines(self):
        return self.lines(self.stderr_head, self.stderr_tail)

    def read_descriptor(self, desc, head, tail, logger, level, h, t, m):
        data = desc.readline()
        if not data:
            return False
        data = data.rstrip('\r\n')
        if logger:
            logger.log(level, data)
        with self.lock:
            if tail:
                tail.append(data)
                if t is not None:
                    tail.pop(0)
                return True
            head.append(data)
            if m is not None and len(head) > m:
                if t != 0:
                    tail.extend(head[-t:])
                del head[h:]
        return True

    def run(self):
        proc = self.proc = Popen(
            self.command,
            stdin=PIPE,
            stdout=PIPE,
            stderr=PIPE,
            # encoding='utf-8',  # This only works on 3.6 and above
            universal_newlines=True,
            start_new_session=True,
            bufsize=1,
            cwd=self.cwd,
        )

        stdout_live, stderr_live = True, True
        while stdout_live or stderr_live:
            r, w, e = select.select([proc.stdout, proc.stderr], [], [])
            if proc.stdout in r:
                stdout_live = self.read_descriptor(
                    proc.stdout,
                    self.stdout_head, self.stdout_tail,
                    self.stdout_logger, self.stdout_level,
                    self.max_lines[0], self.max_lines[1], self.max_lines[2]
                )
            if proc.stderr in r:
                stderr_live = self.read_descriptor(
                    proc.stderr,
                    self.stderr_head, self.stderr_tail,
                    self.stderr_logger, self.stderr_level,
                    self.max_lines[3], self.max_lines[4], self.max_lines[5]
                )

        proc.wait()
        self.returncode = proc.returncode

    def stop(self):
        logger.warning('sending %s to process' % self.stop_signal)
        self.proc.send_signal(self.stop_signal)
        try:
            logger.warning('waiting for process to terminate...')
            self.proc.wait(timeout=self.wait_timeout)
        except TimeoutExpired:
            logger.warning('killing the process...')
            self.proc.kill()
            self.proc.wait()



from email.message import EmailMessage
import ssl
import smtplib
import logging
from threading import Thread


logger = logging.getLogger('periodtask.mailsender')


class MailSender:
    def __init__(
        self, host, port, from_email, recipient_list,
        timeout=10, use_ssl=False, use_tls=False, username=None, password=None,
    ):
        self.host, self.port = host, port
        self.from_email, self.recipient_list = from_email, recipient_list
        self.timeout, self.use_ssl, self.use_tls = timeout, use_ssl, use_tls
        self.username, self.password = username, password

        self.connection_class = smtplib.SMTP_SSL if use_ssl else smtplib.SMTP
        if isinstance(recipient_list, str):
            self.recipient_list = [recipient_list]

    def _send(self, message):
        logger.debug('message sender starts')
        msg = '(%s) to %s' % (message['Subject'], message['To'])
        try:
            conn = self.connection_class(
                self.host, self.port, timeout=self.timeout
            )
            if self.use_tls:
                conn.starttls()
            logger.debug('starttls ok')
            if self.username and self.password:
                conn.login(self.username, self.password)
            logger.debug('login ok')

            try:
                refused = conn.send_message(message)
            except Exception:
                logger.exception('sending error')
            else:
                if refused:
                    logger.warning('refused: %s' % msg)
                else:
                    logger.info('e-mail sent: %s' % msg)

            try:
                conn.quit()
            except (ssl.SSLError, smtplib.SMTPServerDisconnected):
                # This happens when calling quit() on a TLS connection
                # sometimes, or when the connection was already disconnected
                # by the server.
                conn.close()
            except smtplib.SMTPException:
                logger.exception('error during quit')
        except Exception:
            logger.exception('Error sending e-mail: %s' % msg)

    def send_mail(
            self, subject, message, html_message=None
    ):
        msg = EmailMessage()
        msg['Subject'] = subject
        msg['From'] = self.from_email
        msg['To'] = ', '.join(self.recipient_list)
        msg.set_content(message)
        if html_message:
            msg.add_alternative(html_message, subtype='html')
        Thread(target=self._send, args=(msg,)).start()



import logging
import signal
import os

from mako.lookup import TemplateLookup

from .process_thread import ProcessThread
from .periods import Period


logger = logging.getLogger('periodtask.task')
(SKIP, DELAY, RUN) = (0, 1, 2)


base_dir = os.path.dirname(os.path.realpath(__file__))
default_template_dir = os.path.join(base_dir, 'templates')


class Task:
    """
    Represents a task to schedule.

    :param str name: The name of the task, will apear in logs and emails.
    :param tuple command: See ``args`` param of the `Popen constructor
      <https://docs.python.org/3/library/subprocess.html#subprocess.Popen>`_.
    :param list/str periods: A cron expression (str) or a list of them. See
      :doc:`cronref` for more information. By default (when set to an empyt
      string) this will be equivalent to ``0 */5 * * * * UTC``.
    :param bool run_on_start: Indicates weather the task should run when the
      scheduler starts no matter what was given in ``periods``. Useful for
      manually testing the task.
    :param func/bool mail_success: If set to a truthy value an email will be
      sent after the task run successfully. If this is a function, this
      function will be used to send out the email (if **send_mail_func** does
      not override it). The signature of the function is

      .. code-block:: python

        def send_mail(subject, message, html_message=None)
    :param func/bool mail_failure: Controls emails sent when the task fails.
      Otherwise it is the same as **mail_success**.
    :param func/bool mail_skipped: Controls emails sent when the task is
      skipped due to the defined **policy**.
    :param func/bool mail_deleyed: Controls emails sent when the task is
      delayed due to the defined **policy**.
    :param func send_mail_func: If set, this must be a function. This function
      will be used to send emails, no matter what was set in **mail_...**
      params.
    :param number wait_timeout: After sending **stop_signal** to the task
      process, we wait this many seconds for the process to stop. If the
      timeout expires, we kill the process.
    :param int/tuple max_lines: STDOUT and STDERR are collected from the task
      process. To avoid haevy memory usage we only store this many lines in
      memory. More precisely STDOUT head and tail, STDERR head and tail are
      list of lines. This parameter controls the maximum length of these lists.

      examples:

        ==================== ==== ==== ==== ====
        parameter            stdout    stderr
        -------------------- --------- ---------
        value                head tail head tail
        ==================== ==== ==== ==== ====
        ``2``                2    2    2    2
        ``(2, 3)``           2    2    3    3
        ``(10, (2,3))``      10   10   2    3
        ``((1, 2), (3, 4))`` 1    2    3    4
        ==================== ==== ==== ==== ====

    :param int stop_signal: This signal will be sent to the task process when
      we want to stop it gracefully.
    :param int policy: Available values are ``periodtask.SKIP``,
      ``periodtask.DELAY`` and ``periodtask.RUN``.

      **SKIP**
        If a process is (still) running and the task is scheduled, this new
        process will be skipped. If requested, an email will be sent.

      **DELAY**
        If a process is (still) running and the task is scheduled, this new
        process will be delayed and will run immediatelly when the actual
        process terminates. If requested, an email will be sent.

      **RUN**
        Tasks will always run when scheduled.

    :param list/str template_dir: Directories to look for email templates in.
    :param logging.logger stdout_logger: The logger to use for the STDOUT of
      the task process.
    :param int stdout_level: The STDOUT of the task process will be logged to
      this level.
    :param logging.logger stderr_logger: The logger to use for the STDERR of
      the task process.
    :param int stderr_level: The STDERR of the task process will be logged to
      this level.
    :param str cwd: The task process will run with ``cwd`` as the working
      directory. See the `Popen constructor
      <https://docs.python.org/3/library/subprocess.html#subprocess.Popen>`_.
    :param int/None skip_delayed_email_threshold: In havaria situations
      instead of sending SKIP or DELAYED emails forever, only this much
      consecutive emails will be send. When the send queue becomes empty
      a NO BLOCK email will be sent. ``None`` means no threshold.
    :param int/None failure_email_threshold: When a task fails more than
      this in a row no new FAILURE email will be sent. When the task runs
      successfully a RECOVER email will be sent. ``None`` means no threshold.
    """
    def __init__(
        self, name, command,
        periods='',
        run_on_start=False,
        mail_success=None,
        mail_failure=None,
        mail_skipped=None,
        mail_delayed=None,
        send_mail_func=None,
        wait_timeout=10,
        max_lines=50,
        stop_signal=signal.SIGTERM,
        policy=SKIP,
        template_dir=[],
        stdout_logger=logging.getLogger('periodtask.stdout'),
        stdout_level=logging.INFO,
        stderr_logger=logging.getLogger('periodtask.stderr'),
        stderr_level=logging.INFO,
        cwd=None,
        skip_delayed_email_threshold=5,
        failure_email_threshold=5
    ):
        if not isinstance(periods, list) and not isinstance(periods, tuple):
            periods = [periods]
        self.periods = [Period(x) for x in periods]

        self.name = name
        self.command = command
        self.run_on_start = run_on_start
        self.mail_success = mail_success
        self.mail_failure = mail_failure
        self.mail_skipped = mail_skipped
        self.mail_delayed = mail_delayed

        if mail_success and send_mail_func:
            self.mail_success = send_mail_func
        if mail_failure and send_mail_func:
            self.mail_failure = send_mail_func
        if mail_skipped and send_mail_func:
            self.mail_skipped = send_mail_func
        if mail_delayed and send_mail_func:
            self.mail_delayed = send_mail_func

        self.wait_timeout = wait_timeout
        self.max_lines = max_lines
        self.stop_signal = stop_signal
        self.policy = policy
        if isinstance(template_dir, list):
            template_dir = template_dir + [default_template_dir]
        else:
            template_dir = [template_dir] + [default_template_dir]
        self.template_lookup = TemplateLookup(
            directories=template_dir, default_filters=['h']
        )
        self.stdout_logger = stdout_logger
        self.stdout_level = stdout_level
        self.stderr_logger = stderr_logger
        self.stderr_level = stderr_level
        self.cwd = cwd
        self.skip_delayed_email_threshold = skip_delayed_email_threshold
        self.failure_email_threshold = failure_email_threshold

        self.process_threads = []
        self.first_check = True
        self.delay_queue = []
        # self.email_limitation_active = False
        self.failure_email_sent = 0
        self.skip_delayed_email_sent = 0

    def check_second(self, sec):
        if self.first_check:
            self.first_check = False
            if self.run_on_start:
                return 'START'
        for period in self.periods:
            chk = period._check(sec)
            if chk:
                return chk
        return False

    def start_process_thread(self, formatted_sec):
        msg = 'task %s starts process for %s' % (self.name, formatted_sec)
        logger.info(msg)
        thrd = ProcessThread(
            self.name,
            self.command,
            self.stop_signal,
            self.wait_timeout,
            formatted_sec,
            self.max_lines,
            self.stdout_logger,
            self.stdout_level,
            self.stderr_logger,
            self.stderr_level,
            self.cwd,
        )
        self.process_threads.append(thrd)
        thrd.start()

    def send_mail_template(
        self, send_func,
        subject_template, text_template, html_template, **kwargs
    ):
        get_template = self.template_lookup.get_template
        subject = get_template(subject_template)
        subject = subject.render(**kwargs)
        subject = ''.join(subject.splitlines())
        text = get_template(text_template)
        text = text.render(**kwargs)
        html = get_template(html_template)
        html = html.render(**kwargs)
        send_func(subject, text, html_message=html)

    def check_subprocesses(self):
        if not self.process_threads:
            return

        new_process_threads = []
        for subproc in self.process_threads:
            if subproc.is_alive():
                new_process_threads.append(subproc)
                continue
            retcode = subproc.returncode
            msg = 'task %s started for %s terminated with code %s'
            msg = msg % (self.name, subproc.formatted_sec, retcode)
            logger.info(msg)

            if retcode == 0:
                if self.mail_success:
                    self.send_mail_template(
                        self.mail_success,
                        'success_subject.txt',
                        'success.txt',
                        'success.html',
                        subproc=subproc
                    )
                if (
                    self.mail_failure and
                    self.failure_email_threshold is not None and
                    self.failure_email_sent >= self.failure_email_threshold
                ):
                    self.send_mail_template(
                        self.mail_failure,
                        'recover_subject.txt',
                        'recover.txt',
                        'recover.html',
                        task=self,
                        subproc=subproc
                    )
                self.failure_email_sent = 0
            else:
                if (
                    self.mail_failure and (
                        self.failure_email_threshold is None or
                        self.failure_email_sent <
                        self.failure_email_threshold
                    )
                ):
                    self.failure_email_sent += 1
                    self.send_mail_template(
                        self.mail_failure,
                        'failure_subject.txt',
                        'failure.txt',
                        'failure.html',
                        subproc=subproc
                    )

        self.process_threads = new_process_threads

    def skipped_or_delayed(self, formatted_sec, typ='skipped'):
        if getattr(self, 'mail_%s' % typ):
            self.send_mail_template(
                getattr(self, 'mail_%s' % typ),
                '%s_subject.txt' % typ,
                '%s.txt' % typ,
                '%s.html' % typ,
                running=self.process_threads,
                current_sec=formatted_sec,
                task_name=self.name,
                delay_queue=self.delay_queue
            )

    def check_for_second(self, sec):
        formatted_sec = self.check_second(sec)
        if formatted_sec:
            self.delay_queue.append(formatted_sec)

        if self.process_threads:
            if self.policy == SKIP:
                if not formatted_sec:
                    return
                self.delay_queue.pop()  # we skip it
                msg = 'task %s %s for %s' % (
                    self.name, 'skipped', formatted_sec
                )
                logger.warning(msg)

                # do we need to send a skip e-mail?
                if (
                    self.skip_delayed_email_threshold is None or
                    self.skip_delayed_email_sent <
                    self.skip_delayed_email_threshold
                ):
                    self.skipped_or_delayed(formatted_sec)
                    self.skip_delayed_email_sent += 1
                return

            elif self.policy == DELAY:
                if not formatted_sec:
                    return
                msg = 'task %s %s for %s' % (
                    self.name, 'delayed', formatted_sec
                )
                logger.warning(msg)

                # do we need to send a delayed e-mail?
                if (
                    self.skip_delayed_email_threshold is None or
                    self.skip_delayed_email_sent <
                    self.skip_delayed_email_threshold
                ):
                    self.skipped_or_delayed(formatted_sec, typ='delayed')
                    self.skip_delayed_email_sent += 1
                return

        if self.delay_queue:
            self.start_process_thread(self.delay_queue.pop(0))
            if (
                not self.delay_queue and
                self.skip_delayed_email_threshold is not None and
                self.skip_delayed_email_sent >=
                self.skip_delayed_email_threshold
            ):
                func = None
                if self.policy == SKIP:
                    func = self.mail_skipped
                elif self.policy == DELAY:
                    func = self.mail_delayed
                if func:
                    self.send_mail_template(
                        func,
                        'noblock_subject.txt',
                        'noblock.txt',
                        'noblock.html',
                        task=self
                    )
            self.skip_delayed_email_sent = 0
            return True

    def stop(self, check_subprocesses=True):
        if self.process_threads:
            for proc in self.process_threads:
                proc.stop()
                proc.join()
            if check_subprocesses:
                self.check_subprocesses()
        logger.info('task stopped: %s' % self.name)



import logging
import time
import threading
import signal


logger = logging.getLogger('periodtask.tasklist')


class TaskList:
    """
    Defines the tasks to run and starts the sceduler.
    Pass in :py:class:`Task <periodtask.Task>` instances
    to schedule.
    """
    def __init__(self, *args):
        self.tasks = args
        self.last_checked = None
        self.stopped = False
        self.orig_sigint_handler = None
        self.orig_sigterm_handler = None

    def _tick(self):
        now = int(time.time())
        for task in self.tasks:
            task.check_subprocesses()
            for sec in range(self.last_checked + 1, now + 1):
                # Only one process of a task can be started in one tick
                if task.check_for_second(sec):
                    break
        self.last_checked = now

    def start(self):
        """
        Start The scheduler. This will block until ``SIGTERM`` or
        ``SIGINT`` received.
        """
        logger.info('tasklist started')

        def handler(num, frame):
            self._stop()
        self.orig_sigint_handler = signal.getsignal(signal.SIGINT)
        self.orig_sigterm_handler = signal.getsignal(signal.SIGTERM)
        signal.signal(signal.SIGINT, handler)
        signal.signal(signal.SIGTERM, handler)

        self.last_checked = int(time.time()) - 1
        while not self.stopped:
            self._tick()
            time.sleep(0.5)

    def _stop(self, check_subprocesses=True):
        signal.signal(signal.SIGINT, self.orig_sigint_handler)
        signal.signal(signal.SIGTERM, self.orig_sigterm_handler)
        self.stopped = True
        for task in self.tasks:
            task.stop(check_subprocesses)
        for thread in threading.enumerate():
            if thread != threading.main_thread():
                thread.join()



import logging
from datetime import datetime, timedelta

import pytz


logger = logging.getLogger('periodtask.periods')


class BadCronFormat(Exception):
    pass


class Period:
    DOW = {
        'MON': 1, 'TUE': 2, 'WED': 3, 'THU': 4, 'FRI': 5, 'SAT': 6, 'SUN': 7
    }
    SEC_FMT = '{:4}-{:0>2}-{:0>2} {:0>2}:{:0>2}:{:0>2} {}, {}'
    VALID_LF = (
        'F', 'FF', 'FFF', 'FFFF', 'FFFFF',
        'L', 'LL', 'LLL', 'LLLL', 'LLLLL',
    )
    DELTA = timedelta(days=7)

    def __init__(self, cron='0 */5 * * * * UTC'):
        (
            self.seconds, self.minutes, self.hours, self.days,
            self.months, self.years, self.timezone
        ) = self._parse_cron(cron)

    def _parse_cron(self, cron):
        parts = ['0', '*/5', '*', '*', '*', '*', 'UTC']
        try:
            for i, p in enumerate([p for p in cron.split()]):
                parts[i] = p
        except IndexError:
            raise BadCronFormat('too many parts of format string') from None

        seconds = self._parse_part(parts[0], 0, 60)
        minutes = self._parse_part(parts[1], 0, 59)
        hours = self._parse_part(parts[2], 0, 59)
        days = self._parse_part(parts[3], 1, 31, dom_allowed=True)
        months = self._parse_part(parts[4], 1, 12)
        years = self._parse_part(parts[5], 0, None)
        timezone = parts[6]
        try:
            pytz.timezone(timezone)
        except pytz.exceptions.UnknownTimeZoneError:
            raise BadCronFormat('unknown timezone: %s' % timezone) from None

        return (
            seconds, minutes, hours, days, months, years,
            pytz.timezone(timezone)
        )

    def _parse_part(self, part, low, high, dom_allowed=False):
        part = part.strip()
        splitted = part.split(',')
        return [self._parse_split(x, low, high, dom_allowed) for x in splitted]

    def _parse_split(self, split, low, high, dom_allowed):
        try:
            return self._parse_atom(split, low, high, dow=False)
        except BadCronFormat:
            if dom_allowed:
                return self._parse_atom(split, 1, 7, dow=True)
            else:
                raise

    def _parse_atom(self, atom, low, high, dow=False):
        atom = atom.strip()

        # the star is an alias to '-'
        atom = '-' if atom == '*' else atom

        # handle the case when it is a simple integer
        n = self._parse_int(atom, low, high, dow)
        if n is not None:
            return (n, n + 1, 1, dow)

        # parse the step part
        _rng_step = atom.split('/')
        if len(_rng_step) == 1:
            rng, step = atom, 1
        elif len(_rng_step) == 2:
            rng, step = _rng_step
        else:
            desc = 'invalid format: %s' % atom
            raise BadCronFormat(desc)

        # parse the range
        rng = '-' if rng == '*' else rng
        lo_hi = rng.split('-')
        if len(lo_hi) == 1:
            lo, hi = lo_hi[0], lo_hi[0]
        elif len(lo_hi) == 2:
            lo, hi = lo_hi
        else:
            desc = 'invalid format: %s' % atom
            raise BadCronFormat(desc)
        if dow:
            lo, hi = 'MON' if lo == '' else lo, 'SUN' if hi == '' else hi
        else:
            lo, hi = low if lo == '' else lo, high if hi == '' else hi
        lo = self._parse_int(lo, low, high, dow)
        if lo is None:
            desc = 'lower bound is invalid: %s' % atom
            raise BadCronFormat(desc)

        if hi is not None:
            hi = self._parse_int(hi, low, high, dow)
            if hi is None:
                desc = 'upper bound is invalid: %s' % atom
                raise BadCronFormat(desc)

        if hi is not None and lo > hi:
            desc = 'lower bound is too high: %s' % atom
            raise BadCronFormat(desc)

        # parse step
        if dow and step in self.VALID_LF:
            pass
        else:
            step = self._parse_int(step, 1, high)
            if step is None:
                desc = 'step is invalid: %s' % atom
                raise BadCronFormat(desc)
        return (lo, hi if hi is None else hi + 1, step, dow)

    def _parse_int(self, n, low, high, dow=False):
        if dow:
            try:
                return self.DOW[n.upper()]
            except (KeyError, AttributeError):
                return None
        try:
            n = int(n)
        except ValueError:
            return
        else:
            if n < 0:
                return None
            if n < low:
                desc = 'value is lower than %s (%s)' % (low, n)
                raise BadCronFormat(desc)
            if high is not None and n > high:
                desc = 'value is higher than %s (%s)' % (high, n)
                raise BadCronFormat(desc)
            return n

    def _check_part(self, part, actual):
        for lo, hi, step, _ in part:
            hi = actual + 1 if hi is None else hi
            if actual in range(lo, hi, step):
                return True
        return False

    def _check(self, sec):
        utc = datetime.utcfromtimestamp(sec).replace(tzinfo=pytz.utc)
        dt = self.timezone.normalize(utc).astimezone(self.timezone)
        isocalendar = dt.isocalendar()

        year = dt.year
        weekday = isocalendar[2]
        month = dt.month
        day = dt.day
        hour = dt.hour
        minute = dt.minute
        second = dt.second

        if not self._check_part(self.seconds, second):
            return False
        if not self._check_part(self.minutes, minute):
            return False
        if not self._check_part(self.hours, hour):
            return False
        if not self._check_part(self.months, month):
            return False
        if not self._check_part(self.years, year):
            return False

        ok = False
        for lo, hi, step, dow in self.days:
            if not dow:
                if day in range(lo, hi, step):
                    ok = True
                    break
            else:
                if weekday not in range(lo, hi):
                    continue
                if isinstance(step, int):
                    if weekday in range(lo, hi, step):
                        ok = True
                        break
                else:
                    delta = self.DELTA
                    var_dt = dt
                    if step[0] == 'F':
                        delta = -delta
                    ok2 = True
                    for i in range(1, len(step)):
                        var_dt += delta
                        if var_dt.month != month:
                            ok2 = False
                            break
                    if ok2:
                        var_dt += delta
                        if var_dt.month != month:
                            ok = True
                            break
        if not ok:
            return False

        return self.SEC_FMT.format(
                year, month, day, hour, minute, second, self.timezone,
                [x for x in self.DOW if self.DOW[x] == weekday][0]
            )



from setuptools import setup
import os
import io


here = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

VERSION = '0.8.0'

setup(
    name='vertis_periodtask',
    description='Periodic task with timezone',
    long_description=long_description,
    version=VERSION,
    url='https://github.com/vertisfinance/periodtask',
    author='Egyed Gábor',
    author_email='gabor.egyed@vertis.com',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8'
    ],
    install_requires=[
        'pytz >= 2021.1',
        'mako == 1.1.4',
    ],
    license='MIT',
    packages=['periodtask'],
    package_data={
        '': ['templates/*', 'VERSION'],
    }
)
